import asyncio
import logging
import cbor2
import rembus
import websockets
from unittest.mock import patch

payload = 1

mytopic_received = None

async def myservice(data):
    logging.info(f'[myservice]: {data}')
    return data*2

async def mytopic(data):
    global mytopic_received
    logging.info(f'[mytopic]: {data}')
    mytopic_received = payload

class WebSocketMock:
    def __init__(self, responses):
        self.count = 0
        self.responses = responses
        self.queue = asyncio.Queue()

    def build_response(self, msg):
        type_reply = self.responses[self.count][0]
        if type_reply in [rembus.TYPE_RPC, rembus.TYPE_PUB]:
            # pass to application layer
            response_msg = msg
        else:
            msgid = msg[1]
            topic = msg[2]
            #logging.debug(f'[{topic}]: building response')
            sts = self.responses[self.count][1]
            data = self.responses[self.count][2]
            response_msg = [type_reply, msgid, sts, data]
        
        self.count += 1
        return cbor2.dumps(response_msg)
       
    async def send(self, pkt):
        # just to start wait before notifying
        await asyncio.sleep(0.1)
        msg = cbor2.loads(pkt)
        type = msg[0]
        # logging.debug(f'[wsmock] send: type {type} - send [{msg}]')
        await self.queue.put(msg)

    async def close(self):
        pass

    async def recv(self):
        pkt = await self.queue.get()
        # logging.debug(f'[wsmock] recv: {pkt}')
        if pkt[0] == rembus.TYPE_RESPONSE:
            # the message was already processed by rembus handler
            msg = cbor2.dumps(pkt)
        else:
            # build the response for identity, setting, expose, subscribe 
            msg = self.build_response(pkt)
        #logging.debug(f'[wsmock] response: {rembus.tohex(msg)}')
        self.queue.task_done()
        return msg

def ws_mock(mocker, responses):
    return mocker.AsyncMock(return_value=WebSocketMock(responses))


async def test_publish_unknow_topic(mocker):
    topic = "unknown_topic"

    responses = [
        [rembus.TYPE_RESPONSE, rembus.OK, None], # identity
        [rembus.TYPE_RESPONSE, rembus.OK, None], # subscribe
        [rembus.TYPE_PUB]
    ]
    websockets.connect = ws_mock(mocker, responses)

    rb = await rembus.component('foo')
    websockets.connect.assert_called_once_with('ws://localhost:8000/foo', ssl=None)
    logging.info(f'name: {rb.component.name}')
    
    await rb.subscribe(mytopic)
    await rb.publish(topic, payload)
    await rb.close()

async def test_publish(mocker):
    global mytopic_received

    responses = [
        [rembus.TYPE_RESPONSE, rembus.OK, None], # identity
        [rembus.TYPE_RESPONSE, rembus.OK, None], # subscribe
        [rembus.TYPE_PUB],
        [rembus.TYPE_RESPONSE, rembus.OK, None], # unsubscribe
        [rembus.TYPE_PUB],
    ]
    websockets.connect = ws_mock(mocker, responses)

    rb = await rembus.component('foo')
    websockets.connect.assert_called_once_with('ws://localhost:8000/foo', ssl=None)
    assert rb.component.name == 'foo'

    await rb.subscribe(mytopic)
    await rb.publish(mytopic.__name__, payload)
    await asyncio.sleep(0.1)
    assert mytopic_received == payload

    mytopic_received = None
    await rb.unsubscribe(mytopic)
    await rb.publish(mytopic.__name__, payload)

    await asyncio.sleep(0.1)
    assert mytopic_received == None
    await rb.close()

async def test_rpc(mocker):
    responses = [
        [rembus.TYPE_RESPONSE, rembus.OK, None], # identity
        [rembus.TYPE_RESPONSE, rembus.OK, None], # expose
        [rembus.TYPE_RPC], # rpc request
    ]

    websockets.connect = ws_mock(mocker, responses)
    rb = await rembus.component('bar')
    websockets.connect.assert_called_once_with('ws://localhost:8000/bar', ssl=None)
    await rb.expose(myservice)

    response = await rb.rpc(myservice.__name__, payload)
    logging.info(f'response: {response}')
    assert response == payload*2

    await rb.close()

async def test_unexpose(mocker):
    responses = [
        [rembus.TYPE_RESPONSE, rembus.OK, None], # identity
        [rembus.TYPE_RESPONSE, rembus.OK, None], # expose
        [rembus.TYPE_RPC], # rpc request
        [rembus.TYPE_RESPONSE, rembus.OK, None], # unexpose
        [rembus.TYPE_RPC], # rpc request
    ]

    websockets.connect = ws_mock(mocker, responses)
    rb = await rembus.component('bar')
    websockets.connect.assert_called_once_with('ws://localhost:8000/bar', ssl=None)
    await rb.expose(myservice)

    response = await rb.rpc(myservice.__name__, payload)
    logging.info(f'response: {response}')
    assert response == payload*2

    await rb.unexpose(myservice)
    try:
        await rb.rpc(myservice.__name__, payload)
    except Exception as e:
        logging.info(f'unexpose: {e}')

    await rb.close()

async def test_rpc_method_unkown(mocker):
    responses = [
        [rembus.TYPE_RESPONSE, rembus.OK, None], # identity
        [rembus.TYPE_RESPONSE, rembus.OK, None], # expose
        [rembus.TYPE_RPC], # rpc request
    ]

    websockets.connect = ws_mock(mocker, responses)
    rb = await rembus.component('bar')
    websockets.connect.assert_called_once_with('ws://localhost:8000/bar', ssl=None)
    await rb.expose(myservice)

    invalid_method = 'invalid_method'
    try:
        response = await rb.rpc(invalid_method, payload)
    except Exception as e:
        logging.info(e.message)
        assert isinstance(e, rembus.RembusError)
        assert e.status == rembus.METHOD_NOT_FOUND
        assert e.message == invalid_method
    await rb.close()

async def test_rpc_method_exception(mocker):
    responses = [
        [rembus.TYPE_RESPONSE, rembus.OK, None], # identity
        [rembus.TYPE_RESPONSE, rembus.OK, None], # expose
        [rembus.TYPE_RPC], # rpc request
    ]

    websockets.connect = ws_mock(mocker, responses)
    rb = await rembus.component('bar')
    websockets.connect.assert_called_once_with('ws://localhost:8000/bar', ssl=None)
    await rb.expose(myservice)

    try:
        await rb.rpc(myservice.__name__)
    except Exception as e:
        logging.info(e)
        assert isinstance(e, rembus.RembusError)
        assert e.status == rembus.METHOD_EXCEPTION
        assert e.message == "myservice() missing 1 required positional argument: 'data'"
    await rb.close()   


import cbor2
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import serialization
import logging
import os
import pandas as pd
from platformdirs import *
import pyarrow as pa
import ssl
from urllib.parse import urlparse
import uuid
import websockets

from .protocol import *

# Configure the logger
logging.basicConfig(level=logging.ERROR)

# create logger
logger = logging.getLogger("rembus")

def tohex(bytes):
    """Return a string with bytes as hex numbers with 0xNN format."""
    return ' '.join(f'0x{x:02x}' for x in bytes)


def field_repr(bstr):
    """String repr of the second field of a rembus message.

    The second field may be a 16-bytes message unique id or a topic string
    value.  
    """
    return tohex(bstr) if isinstance(bstr, bytes) else bstr


def msg_str(dir, msg):
    """Return a printable dump of rembus message `msg`."""
    payload = ", ".join(str(el) for el in msg[2:])
    s = f'{dir}: [{msg[0]}, {field_repr(msg[1])}, {payload}]'
    return s

class Component:

    def __init__(self, url=None) -> None:
        uri = urlparse(url)
        if uri.path:
            self.name = uri.path[1:] if uri.path.startswith("/") else uri.path
        else:
            self.name = None
        if uri.scheme:
            self.scheme = uri.scheme
        else:
            self.scheme = "ws"
        if uri.netloc:
            self.netloc = uri.netloc
        else:
            self.netloc = os.getenv('REMBUS_BASE_URL', 'localhost:8000')

    def connection_url(self):
        if self.name:
            return f"{self.scheme}://{self.netloc}/{self.name}"
        else:
            return f"{self.scheme}://{self.netloc}"


def decode_dataframe(data):
    """Decode a CBOR tagged value `data` to a pandas dataframe."""
    writer = pa.BufferOutputStream()
    writer.write(data)
    buf = writer.getvalue()
    reader = pa.ipc.open_stream(buf)
    with pa.ipc.open_stream(buf) as reader:
        return reader.read_pandas()


def encode_dataframe(df):
    """Encode a pandas dataframe `df` to a CBOR tag value."""
    table = pa.Table.from_pandas(df)
    sink = pa.BufferOutputStream()
    with pa.ipc.new_stream(sink, table.schema) as writer:
        writer.write(table)
    buf = sink.getvalue()
    stream = pa.input_stream(buf)
    return cbor2.CBORTag(DATAFRAME_TAG, stream.read())


def encode(msg):
    """Encode message `msg`."""
    logger.debug(msg_str('out', msg))
    return cbor2.dumps(msg)


def tag2df(data):
    """Loop over `data` items and decode tagged values to dataframes."""
    if isinstance(data, list):
        for idx, val in enumerate(data):
            if isinstance(val, cbor2.CBORTag) and val.tag == DATAFRAME_TAG:
                data[idx] = decode_dataframe(val.value)
    elif isinstance(data, cbor2.CBORTag):
        return decode_dataframe(data.value)
    return data


def df2tag(data):
    """Loop over `data` items and encode dataframes to tag values."""
    if isinstance(data, tuple):
        lst = []
        for idx, val in enumerate(data):
            if isinstance(val, pd.DataFrame):
                lst.append(encode_dataframe(val))
            else:
                lst.append(val)
        return lst
    elif isinstance(data, list):
        for idx, val in enumerate(data):
            if isinstance(val, pd.DataFrame):
                data[idx] = encode_dataframe(val)

    elif isinstance(data, pd.DataFrame):
        data = encode_dataframe(data)
    return data


def id():
    """Return an array of 16 random bytes."""
    return uuid.uuid4().bytes


def config_dir():
    """The directory for rembus secrets."""
    # appears in path only for Windows machine
    app_author = "Rembus"
    return user_config_dir("rembus", app_author)


def save_private_key(cid, private_key):
    dir = config_dir()
    if not os.path.exists(dir):
        os.makedirs(dir)

    fn = os.path.join(dir, cid)
    private_key_file = open(fn, "wb")
    private_key_file.write(private_key)
    private_key_file.close()


def load_private_key(cid):
    fn = os.path.join(config_dir(), cid)
    with open(fn, "rb") as key_file:
        private_key = serialization.load_pem_private_key(
            key_file.read(), password=None)

    return private_key



import os

WS_FRAME_MAXSIZE = 60 * 1024 * 1024

TYPE_IDENTITY = 0
TYPE_PUB = 1
TYPE_RPC = 2
TYPE_ADMIN = 3
TYPE_RESPONSE = 4
TYPE_ACK = 6
TYPE_REGISTER = 10
TYPE_ATTESTATION = 11

OK = 0
ERROR = 0x0A
CHALLENGE = 0x0B            # 11
IDENTIFICATION_ERROR = 0X14 # 20
METHOD_EXCEPTION = 0X28     # 40
METHOD_ARGS_ERROR = 0X29    # 41
METHOD_NOT_FOUND = 0X2A     # 42
METHOD_UNAVAILABLE = 0X2B   # 43
METHOD_LOOPBACK = 0X2C      # 44
TARGET_NOT_FOUND = 0X2D     # 45
TARGET_DOWN = 0X2E          # 46
UNKNOWN_ADMIN_CMD = 0X2F    # 47

DATAFRAME_TAG = 80

retcode = {
    OK: 'OK',
    ERROR: 'Unknown error',
    IDENTIFICATION_ERROR: 'IDENTIFICATION_ERROR',
    METHOD_EXCEPTION: 'METHOD_EXCEPTION',
    METHOD_ARGS_ERROR: 'METHOD_ARGS_ERROR',
    METHOD_NOT_FOUND: 'METHOD_NOT_FOUND',
    METHOD_UNAVAILABLE: 'METHOD_UNAVAILABLE',
    METHOD_LOOPBACK: 'METHOD_LOOPBACK',
    TARGET_NOT_FOUND: 'TARGET_NOT_FOUND',
    TARGET_DOWN: 'TARGET_DOWN',
    UNKNOWN_ADMIN_CMD: 'UNKNOWN_ADMIN_CMD',
}

BROKER_CONFIG = '__config__'
COMMAND = 'cmd'
ADD_INTEREST = 'subscribe'
REMOVE_INTEREST = 'unsubscribe'
ADD_IMPL = 'expose'
REMOVE_IMPL = 'unexpose'

class RembusException(Exception):
    pass

class RembusTimeout(RembusException):
    def __str__(self):
        return 'request timeout'

class RembusConnectionClosed(RembusException):
    def __str__(self):
        return 'connection closed'

class RembusError(RembusException):
    def __init__(self, status_code, msg=None):
        self.status = status_code
        self.message = msg

    def __str__(self):
        if self.message:
            return f'{retcode[self.status]}:{self.message}'
        else:
            return f'{retcode[self.status]}'

def request_timeout():
    return float(os.environ.get('REMBUS_TIMEOUT', 10.0))


def msg_status(response):
    return response[2]

def msg_id(response):
    return response[1]




import asyncio
from async_timeout import timeout
from rembus.common import *

logger = logging.getLogger("rembus")

background_tasks = set()

async def component_task(cmp):
    while True:
        msg = await cmp.inbox.get()
        logger.debug(f"component task: {msg}")
        if msg == "shutdown":
            break

        logger.debug(f"{cmp.name}: reconnecting ...")
        try:
            await cmp.connect()
            await cmp.reactive()
        except Exception as e:
            logger.error(f"{cmp.name} connect: {e}")
            await asyncio.sleep(2)
            cmp.inbox.put_nowait("reconnect")

async def component(name=None):
    if name in connected_components:
        return connected_components[name]
    else:
        cmp = Rembus(name)
        await cmp.connect()
        add_component(name, cmp)
        cmp.task = asyncio.create_task(component_task(cmp))
        cmp.task.add_done_callback(background_tasks.discard)
        return cmp

class Rembus:
    def __init__(self, name=None):
        self.name = name
        self.ws = None
        self.receiver = None
        self.component = Component(name)
        self.inbox = asyncio.Queue()

        # outstanding requests
        self.outreq = {}
        self.handler = {}

    async def evaluate(self, topic, data):
        """Invoke the handler associate with the message topic.

        :meta private:
        """
        if isinstance(data, list):
            output = await self.handler[topic](*data)
        elif isinstance(data, bytes):
            args = list(data)
            output = await self.handler[topic](*args)
        else:
            output = await self.handler[topic](data)
        return output

    async def parse_input(self, msg):
        """:meta private:"""
        type_byte, msgid = msg[0:2]

        type = type_byte & 0x3F
        flags = type_byte & 0xC0
        logger.debug(f"recv packet type {type}, flags:{flags}")

        if type == TYPE_PUB:
            data = tag2df(msg[2])
            try:
                await self.evaluate(msgid, data)
            except Exception as e:
                logger.error(f"{e}")
            return
        elif type == TYPE_RPC:
            data = tag2df(msg[4])
            topic = msg[2]

            if not topic in self.handler:
                outmsg = [TYPE_RESPONSE, msgid, METHOD_NOT_FOUND, topic]
            else:
                status = OK
                try:
                    output = await self.evaluate(topic, data)
                except Exception as e:
                    status = METHOD_EXCEPTION
                    output = f"{e}"
                    logger.info(f"exception: {e}")

                outmsg = [TYPE_RESPONSE, msgid, status, df2tag(output)]
                logger.debug(msg_str('out', outmsg))

            await self.ws.send(cbor2.dumps(outmsg))
            return

        fut = self.outreq.pop(msgid, None)
        if fut == None:
            logger.warning(f"recv unknown msg id {tohex(msgid)}")
            return

        if type == TYPE_RESPONSE:
            sts = msg[2]
            payload = (msg[3:] + [None])[0]
            if sts == OK:
                fut.set_result(tag2df(payload))
            elif sts == CHALLENGE:
                fut.set_result(payload)
            else:
                fut.set_exception(RembusError(sts, payload))

    async def receive(self):
        """:meta private:"""
        try:
            while True:
                result = await self.ws.recv()
                msg = cbor2.loads(result)
                logger.debug(msg_str('in', msg))
                await self.parse_input(msg)
        except websockets.ConnectionClosedOK:
            logger.debug("connection closed")
        except Exception as e:
            logger.warning(f"closing: {e}")
        finally:

            # for now send a message to task
            await self.inbox.put("reconnect")

            await self.close()

    async def connect(self):
        """Connect to the broker."""
        broker_url = self.component.connection_url()

        ssl_context = None
        if self.component.scheme == "wss":
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ca_crt = os.getenv("HTTP_CA_BUNDLE", "rembus-ca.crt")
            if os.path.isfile(ca_crt):
                ssl_context.load_verify_locations(ca_crt)
            else:
                logger.warning(f"CA file not found: {ca_crt}")

        self.ws = await websockets.connect(broker_url, ssl=ssl_context)
        self.receiver = asyncio.create_task(self.receive())

        if self.component.name != None:
            try:
                await self.login()
            except Exception as e:
                raise RembusError("login failed")
        return self

    async def send_wait(self, builder):
        """:meta private:"""
        reqid = id()
        await self.ws.send(builder(reqid))
        self.outreq[reqid] = asyncio.get_running_loop().create_future()
        try:
            async with timeout(3):
                return await self.outreq[reqid]
        except TimeoutError:
            raise RembusTimeout()

    async def login(self):
        """:meta private:"""
        challenge = await self.send_wait(
            lambda id: encode([TYPE_IDENTITY, id, self.component.name])
        )
        if challenge and isinstance(challenge, bytes):
            logger.debug(f"challenge: {challenge}")
            plain = [bytes(challenge), self.component.name]
            message = cbor2.dumps(plain)
            logger.debug(f"message: {message.hex()}")
            self.privatekey = load_private_key(self.component.name)
            signature = self.privatekey.sign(
                message, padding.PKCS1v15(), hashes.SHA256()
            )
            await self.send_wait(
                lambda id: encode(
                    [TYPE_ATTESTATION, id, self.component.name, signature])
            )
        else:
            logger.debug(f"cid {self.component.name}: free mode access")

    async def publish(self, topic, *args):
        data = df2tag(args)
        await self.ws.send(encode([TYPE_PUB, topic, data]))

    async def broker_setting(self, command, args={}):
        data = {COMMAND: command} | args
        return await self.send_wait(
            lambda id: encode([TYPE_ADMIN, id, BROKER_CONFIG, data])
        )

    async def setting(self, topic, command, args={}):
        data = {COMMAND: command} | args
        return await self.send_wait(lambda id: encode([TYPE_ADMIN, id, topic, data]))

    async def rpc(self, topic, *args):
        data = df2tag(args)
        return await self.send_wait(
            lambda id: encode([TYPE_RPC, id, topic, None, data])
        )

    async def direct(self, target, topic, *args):
        data = df2tag(args)
        return await self.send_wait(
            lambda id: encode([TYPE_RPC, id, topic, target, data])
        )

    async def reactive(self):
        await self.broker_setting("reactive", {"status": True})
        return self

    async def unreactive(self):
        await self.broker_setting("reactive", {"status": False})
        return self

    async def subscribe(self, fn, retroactive=False):
        topic = fn.__name__
        await self.setting(topic, ADD_INTEREST, {"retroactive": retroactive})
        self.handler[topic] = fn
        return self

    async def unsubscribe(self, fn):
        if isinstance(fn, str):
            topic = fn
        else:
            topic = fn.__name__

        await self.setting(topic, REMOVE_INTEREST)
        self.handler.pop(topic, None)
        return self

    async def expose(self, fn):
        topic = fn.__name__
        self.handler[topic] = fn
        await self.setting(topic, ADD_IMPL)

    async def unexpose(self, fn):
        if isinstance(fn, str):
            topic = fn
        else:
            topic = fn.__name__

        self.handler.pop(topic, None)
        await self.setting(topic, REMOVE_IMPL)

    async def shutdown(self):
        await self.inbox.put("shutdown")
        
    async def close(self):
        self.receiver.cancel()
        if self.ws:
            await self.ws.close()
            self.ws = None

    async def forever(self):
        await self.reactive()
        await self.task



import functools
import threading
import time
from websockets.sync.client import connect
from rembus.common import *

# Global dictionary to store connected components
connected_components = {}

def add_component(name, component):
    """Add a component to the connected components dictionary."""
    connected_components[name] = component

def get_component(name):
    """Retrieve a component from the connected components dictionary."""
    return connected_components.get(name)

def remove_component(name):
    """Remove a component from the connected components dictionary."""
    if name in connected_components:
        del connected_components[name]

def connector(cmp):
    retries = 0
    while True:
        if cmp.ws:
            with cmp.isdone:
                cmp.isdone.wait()

        if cmp.isclosing():
            break

        time.sleep(4)
        logger.debug(f"{cmp.name}: reconnecting ...")
        cmp.connect()
        retries += 1

def component(name=None):
    if name in connected_components:
        return connected_components[name]
    else:
        cmp = Rembus(name)
        cmp.connect()
        add_component(name, cmp)
    
        conn = threading.Thread(
            target=connector, args=(cmp,),
            daemon=False)
        conn.start()
        return cmp

class Rembus:
    def __init__(self, name=None):
        self.name = name
        self.status = 'IDLE'
        self.ws = None
        self.receiver = None
        self.component = Component(name)

        # outstanding requests
        self.outreq = {}
        self.handler = {}

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def evaluate(self, topic, data):
        """:meta private:"""
        if isinstance(data, list):
            output = self.handler[topic](*data)
        elif isinstance(data, bytes):
            args = list(data)
            output = self.handler[topic](*args)
        else:
            output = self.handler[topic](data)
        return output

    def parse_input(self, msg):
        """:meta private:"""
        type_byte, msgid = msg[0:2]

        type = type_byte & 0x3F
        flags = type_byte & 0xC0
        logger.debug(f"recv packet type {type}, flags:{flags}")
        if type == TYPE_PUB:
            data = tag2df(msg[2])
            try:
                self.evaluate(msgid, data)
            except Exception as e:
                logger.error(f"{e}")
            return
        elif type == TYPE_RPC:
            data = tag2df(msg[4])
            topic = msg[2]

            if not topic in self.handler:
                outmsg = [TYPE_RESPONSE, msgid, METHOD_NOT_FOUND, topic]
            else:
                status = OK
                try:
                    output = self.evaluate(topic, data)
                except Exception as e:
                    status = METHOD_EXCEPTION
                    output = f"{e}"
                    logger.info(f"exception: {e}")

                outmsg = [TYPE_RESPONSE, msgid, status, df2tag(output)]
                logger.debug(msg_str('out', outmsg))

            self.ws.send(cbor2.dumps(outmsg))
            return

        condition = self.outreq.pop(msgid, None)
        if condition == None:
            logger.warning(f"recv unknown msg id {tohex(msgid)}")
            return
        with condition:
            if type == TYPE_RESPONSE:
                sts = msg[2]
                if sts == OK:
                    self.outreq[msgid] = tag2df(msg[3])
                elif sts == CHALLENGE:
                    self.outreq[msgid] = msg[3]
                else:
                    self.outreq[msgid] = RembusError(sts, msg[3])

            condition.notify()

    def receive(self):
        """:meta private:"""
        try:
            while True:
                result = self.ws.recv()
                msg = cbor2.loads(result)
                logger.debug(msg_str('in', msg))
                self.parse_input(msg)
        except websockets.exceptions.ConnectionClosedError:
            logger.info("unexpected ws close connection")
        except Exception as e:
            logger.info(f"closing ({type(e)}): {e}")
        finally:
            self.close_connection()

            with self.isdone:
                self.isdone.notify()
            
            # notify all outstanding requests
            for reqid in self.outreq:
                cond = self.outreq[reqid]
                if isinstance(cond, threading.Condition):
                    with cond:
                        self.outreq[reqid] = RembusConnectionClosed()
                        cond.notify()

    def connect(self):
        self.isdone = threading.Condition()
        broker_url = self.component.connection_url()

        ssl_context = None
        if self.component.scheme == "wss":
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ca_crt = os.getenv("HTTP_CA_BUNDLE", "rembus-ca.crt")
            if os.path.isfile(ca_crt):
                ssl_context.load_verify_locations(ca_crt)
            else:
                logger.warning(f"CA file not found: {ca_crt}")

        try:
            self.ws = connect(broker_url, max_size=WS_FRAME_MAXSIZE, ssl_context=ssl_context)
            self.receiver = threading.Thread(
                target=self.receive, args=(), daemon=False)
            self.receiver.start()
            if self.component.name != None:
                try:
                    self.login()
                except Exception as e:
                    raise RembusError(IDENTIFICATION_ERROR)
            self.status = 'CONNECTED'
        except ConnectionRefusedError:
            with self.isdone:
                self.isdone.notify()
        return self

    def timeout(self, reqid):
        cond = self.outreq[reqid]
        if isinstance(cond, threading.Condition):
            with cond:
                self.outreq[reqid] = RembusTimeout()
                cond.notify()

    def send_wait(self, builder):
        """:meta private:"""
        reqid = id()
        condition = threading.Condition()
        self.outreq[reqid] = condition
        self.ws.send(builder(reqid))

        timer = threading.Timer(
            request_timeout(), lambda reqid: self.timeout(reqid), args=(reqid,)
        )
        timer.start()

        with condition:
            condition.wait()
        timer.cancel()
        result = self.outreq.pop(reqid)
        if isinstance(result, RembusException):
            raise result
        
        return result

    def login(self):
        """:meta private:"""
        challenge = self.send_wait(
            lambda id: encode([TYPE_IDENTITY, id, self.component.name])
        )
        if challenge and isinstance(challenge, bytes):
            logger.debug(f"challenge: {challenge}")
            plain = [bytes(challenge), self.component.name]
            message = cbor2.dumps(plain)
            self.privatekey = load_private_key(self.component.name)
            signature = self.privatekey.sign(
                message, padding.PKCS1v15(), hashes.SHA256()
            )
            self.send_wait(
                lambda id: encode(
                    [TYPE_ATTESTATION, id, self.component.name, signature])
            )
        else:
            logger.debug(f"cid {self.component.name}: free mode access")

    def publish(self, topic, *args):
        data = df2tag(args)
        self.ws.send(encode([TYPE_PUB, topic, data]))

    def broker_setting(self, command, args={}):
        data = {COMMAND: command} | args
        return self.send_wait(
            lambda id: encode([TYPE_ADMIN, id, BROKER_CONFIG, data])
        )

    def setting(self, topic, command, args={}):
        data = {COMMAND: command} | args
        return self.send_wait(lambda id: encode([TYPE_ADMIN, id, topic, data]))

    def rpc(self, topic, *args):
        data = df2tag(args)
        return self.send_wait(
            lambda id: encode([TYPE_RPC, id, topic, None, data]))

    def direct(self, target, topic, *args):
        data = df2tag(args)
        return self.send_wait(
            lambda
            id: encode(
                [TYPE_RPC, id, topic, target, data]))

    def reactive(self):
        self.broker_setting("reactive", {"status": True})
        return self

    def unreactive(self):
        self.broker_setting("reactive", {"status": False})
        return self

    def subscribe(self, fn, retroactive=False):
        topic = fn.__name__
        self.setting(topic, ADD_INTEREST, {"retroactive": retroactive})
        self.handler[topic] = fn
        return self

    def unsubscribe(self, fn):
        if isinstance(fn, str):
            topic = fn
        else:
            topic = fn.__name__

        self.setting(topic, REMOVE_INTEREST)
        self.handler.pop(topic, None)
        return self

    def expose(self, fn):
        topic = fn.__name__
        self.handler[topic] = fn
        self.setting(topic, ADD_IMPL)

    def unexpose(self, fn):
        if isinstance(fn, str):
            topic = fn
        else:
            topic = fn.__name__

        self.handler.pop(topic, None)
        self.setting(topic, REMOVE_IMPL)

    def close_connection(self):
        if not self.isclosing():
            self.status = 'DISCONNECTED'

        if self.ws:
            self.ws.close()
            self.ws = None

    def close(self):
        self.status = 'CLOSED'
        remove_component(self.name)
        self.close_connection()

    def isclosing(self):
        return self.status == 'CLOSED'
    
    def isopen(self):
        return self.status == 'CONNECTED'

    def forever(self):
        self.reactive()
        self.receiver.join()
